/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 * 
 * TCP Proxy skeleton.
 * 
 */
#include "async.h"

#define MAX_LISTEN_BACKLOG 5
#define MAX_ADDR_NAME	32
#define ONE_K	1024

/* Data that can pile up to be sent before reading must stop temporarily */
#define MAX_CONN_BACKLOG	(8*ONE_K)
#define GRACE_CONN_BACKLOG	(MAX_CONN_BACKLOG / 2)

/* Watermarks for number of active connections. Lower to 2 for testing */
#define MAX_CONN_HIGH_WATERMARK	(256)
#define MAX_CONN_LOW_WATERMARK	(MAX_CONN_HIGH_WATERMARK - 1)


struct sockaddr_in remote_addr; /* The address of the target server */
static int proxy_sock; /* The descriptor of the socket the proxy listens on */

/* called by the event loop when ready to read / accept connected socket */
static void connect_to_proxy(void)
{
	int client_sock, server_sock;
	struct sockaddr_in proxy_addr, client_addr;
	socklen_t addr_len = sizeof(struct sockaddr_in);
	char client_hname[MAX_ADDR_NAME+1], server_hname[MAX_ADDR_NAME+1];
	
	memset(&proxy_addr, 0, sizeof(struct sockaddr_in));
	
	if ((client_sock = accept(proxy_sock, (struct sockaddr *) &proxy_addr, 
		&addr_len)) < 0)
	{
		if (errno != EAGAIN) /* nonblocking, retry */
			fprintf(stderr, "accept error %s\n", strerror(errno));
	}
	else
	{
		make_async(client_sock); /* this may very well be redundant */
		if ((server_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			fprintf(stderr, "socket error %s\n", strerror(errno));
			close(client_sock);
		}
		make_async(server_sock);

		if (connect(server_sock, (struct sockaddr *) &remote_addr, 
			sizeof(struct sockaddr_in)) < 0)
		{
			if (errno != EINPROGRESS) /* nonblocking, connection stalled */
			{
				fprintf(stderr, "connect error %s\n", strerror(errno));
				close(client_sock);
				close(server_sock);
			}
		}
		
		addr_len = sizeof(struct sockaddr_in); /* in/out parameter */
		if (getpeername(client_sock, (struct sockaddr *) &client_addr, 
				&addr_len) < 0)
		{
			fprintf(stderr, "getsockname error %s\n", strerror(errno));
			exit(1);
		}

		strncpy(client_hname, inet_ntoa(client_addr.sin_addr), MAX_ADDR_NAME);
		strncpy(server_hname, inet_ntoa(remote_addr.sin_addr), MAX_ADDR_NAME);
		printf("Connection proxied: %s:%d --> %s:%d\n", 
				client_hname, ntohs(client_addr.sin_port), 
				server_hname, ntohs(remote_addr.sin_port));
		
		close(server_sock);
		close(client_sock);
	}
	
	/* Enforce the watermark of maximum proxy connections: this can be done by
	 * removing the callback for the proxy_sock descriptor from the event loop:
	 *				
	 *				fdcb(proxy_sock, selread, NULL);
	 * 
	 * At the same time, it must be re-enabled when the number of active proxy
	 * connections drops below the low watermark. 
	 */
}

int main(int argc, char **argv)
{
	char *remote_name;
	struct sockaddr_in proxy_addr;
	unsigned short local_port, remote_port;
	struct hostent *h;
	int arg_idx = 1;
	
	if (argc != 4)
	{
		fprintf(stderr, "Usage %s <remote-target> <remote-target-port> "
			"<local-port>\n", argv[0]);
		exit(1);
	}

	remote_name = argv[arg_idx++];
	remote_port = atoi(argv[arg_idx++]);
	local_port = atoi(argv[arg_idx++]);
	
	/* Lookup server name and establish control connection */
	if ((h = gethostbyname(remote_name)) == NULL)
	{
		fprintf(stderr, "gethostbyname(%s) failed %s\n", remote_name, 
			strerror(errno));
		exit(1);
	}
	memset(&remote_addr, 0, sizeof(struct sockaddr_in));
	remote_addr.sin_family = AF_INET;
	memcpy(&remote_addr.sin_addr.s_addr, h->h_addr_list[0], sizeof(in_addr_t));
	remote_addr.sin_port = htons(remote_port);
	
	/* open up the TCP socket the proxy listens on */
	if ((proxy_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
	{
		fprintf(stderr, "socket error %s\n", strerror(errno));
		exit(1);
	}
	/* bind the socket to all local addresses */
	memset(&proxy_addr, 0, sizeof(struct sockaddr_in));
	proxy_addr.sin_family = AF_INET;
	proxy_addr.sin_addr.s_addr = INADDR_ANY; /* bind to all local addresses */
	proxy_addr.sin_port = htons(local_port);
	if (bind(proxy_sock, (struct sockaddr *) &proxy_addr, 
			sizeof(proxy_addr)) < 0)
	{
		fprintf(stderr, "bind error %s\n", strerror(errno));
		exit(1);
	}
	
	/* make the socket async */
	make_async(proxy_sock);
	listen(proxy_sock, MAX_LISTEN_BACKLOG);
	/* register socket and signal handlers with the async event loop */
	fdcb(proxy_sock, selread, wrap(connect_to_proxy));
	sigcb(SIGINT, wrap(exit, 1));
	sigcb(SIGINT, wrap(exit, 1));
	
	amain(); /* async main */
}

